/*
 * ClauseStorage.hpp
 *
 *  Created on: Aug 18, 2009
 *      Author: kottlers
 */

#ifndef CLAUSESTORAGE_HPP_
#define CLAUSESTORAGE_HPP_
#define clauseinfos_hpp_version      "$Revision: 1.14 $"

#include "GlobalSettings.hpp"
#include "AtomicTypes.hpp"
#include "BitVector.hpp"
#include "SharedClause.hpp"
#include "Satistics.hpp"


class ClsPointer{
	uint32_t ID;
public:
	static const uint32_t INVALID_IDX 	= max32;
	//static const uint32_t NEARLY_NULL	= max32-1;
	explicit ClsPointer(uint32_t i):ID(i){};
	uint32_t idx() const	{return ID;}

	ClsPointer():ID(INVALID_IDX){};

	bool isNull()	const	{return ID == INVALID_IDX;}
	void setNull()			{ID = INVALID_IDX;}

	//just used to mark return values where NULL has a different meaning
	//bool nearlyNull()const	{return ID == NEARLY_NULL;}
	//bool anyNull()const		{return ID >= NEARLY_NULL;}

	friend bool operator==(const ClsPointer& p1,const ClsPointer& p2);
	friend bool operator!=(const ClsPointer& p1,const ClsPointer& p2);
};

class ClsWrp{
	friend class ClsInfo;
	uint32_t wtches;
	uint32_t do_bup : 1;
	uint32_t lockd  : 1;
	uint32_t Lbd	: 4;
	uint32_t offset	: 26;
	SharedClause*  cls;
public:
	Literal getOtherW( const Literal& l)const{
		return Literal(l.idx() ^ wtches);}

	void	setWatchers(const Literal& l1,const Literal& l2){ asrt(l1!=l2);
		wtches = l1.idx() ^ l2.idx();}

	void	invalidateWatchers(){wtches= 0;}
	bool	watchersInvalid(){	return wtches == 0;}

	SharedClause* getCls()	{return cls;}
	void silentReplace(SharedClause* newc, const uint32_t& lbd){cls = newc; Lbd=lbd;}
	bool shallDoBUP()const	{return do_bup;}
	void setBUP(bool b)		{do_bup=b;}
	void setLockd(bool b)	{lockd=b;}
	bool isLockd()	const	{return lockd;}
	uint32_t getLBD() const	{asrt(Lbd==0 || cls->isLearnt()); return Lbd;}
	void setLBD(uint32_t LBD){Lbd=LBD; asrt(LBD <= SharedClause::LBD_MAX); asrt(cls->isLearnt());}

	ClsLit* addOffset(ClsLit* strt)const{return strt + offset;}
	void keepOffset(const ClsLit* strt, const ClsLit* found){asrt(found >= strt); offset = found-strt;}
	void setOffset(uint32_t sz)	{offset = sz; asrt(offset == sz);}
	uint32_t getOffset()		{return offset;}

};


class ClsInfo{

	einfo<ClsWrp,ClsPointer>			cdata;
	ClsPointer 							free_data_head; //list of available pointers
	Satistics*							stats;


	inline ClsPointer	nextValidFrom(uint32_t idx) const;
	inline ClsPointer	nextValidBefore(uint32_t idx) const;

	void	count(const SharedClause* cls,const uint32_t sz){asrt(cls!=NULL);
			if(cls->isLearnt()){++stats->nbr_ll_cls; stats->nbr_ll_lits+=sz; }
			else{++stats->nbr_lf_cls; stats->nbr_lf_lits+=sz;}}
	void	uncount(const SharedClause* cls,const uint32_t sz){asrt(cls!=NULL);
			if(cls->isLearnt()){asrt(stats->nbr_ll_cls!=0);--stats->nbr_ll_cls; asrt(stats->nbr_ll_lits>=sz); stats->nbr_ll_lits-=sz;}
			else{asrt(stats->nbr_lf_cls!=0); --stats->nbr_lf_cls; asrt(stats->nbr_lf_lits>=sz); stats->nbr_lf_lits-=sz;}}
public:

	ClsInfo(Satistics* st,uint32_t sz=1500):free_data_head(ClsPointer::INVALID_IDX),stats(st){allocate(sz);}
	void createClone(const ClsInfo& clone){
		cdata		= clone.cdata;
	}

	uint32_t clsCount() const {return cdata.size();}
	bool 	validPtr(const ClsPointer& p)const { return cdata.inRange(p) && (cdata[p].cls != NULL);}
	bool 	validNoWaste(const ClsPointer& p)const {
		return cdata.inRange(p) && (cdata[p].cls != NULL) && (!cdata[p].cls->isWaste());}

	void allocate(uint32_t sz){
		cdata.reserve(sz);}

	//ClsPointer	nextValidBehind(const ClsPointer& p)	{return nextValidFrom(p.idx() + 1);}
	ClsPointer	nextValidBefore(const ClsPointer& p)const	{return nextValidBefore(p.idx());}

	//ClsPointer	firstValid()							{return nextValidFrom(0);}
	ClsPointer	lastValid() const						{return nextValidBefore(cdata.size());}

	//inline void	loadUpdate(const ClsPointer& p,const UserBit& ub);

	inline ClsPointer addCls(const Literal& w1,const Literal& w2, SharedClause* cls,const uint32_t sz, const uint32_t& lbd);

	//inline void	freeCls(const ClsPointer& p, const UserBit& ub,string why);
	inline void	freeCls(const ClsPointer& p, const UserBit& ub);
	inline void	freePtrOnly(const ClsPointer& p, const uint32_t sz);

	ClsWrp&  get(const ClsPointer& clp){asrt(clp.idx()<cdata.size());return cdata[clp];}
	ClsWrp   getCopy(const ClsPointer& clp)const {return cdata[clp];}
	SharedClause* getCls(const ClsPointer& clp){return get(clp).cls;}
	bool	isLearnt(const ClsPointer& clp)const {return cdata[clp].cls->isLearnt();}

	void	replaceShared(const ClsPointer& clp, SharedClause* new_one, int32_t cnt ){
			ClsWrp& cw = get(clp); asrt(cw.cls->isLearnt() == new_one->isLearnt());
			cw.cls = new_one;
			*(cw.cls->isLearnt() ? &stats->nbr_ll_lits : & stats->nbr_lf_lits) += cnt;}

	inline uint32_t getLBDWithRank(const uint32_t rank);
};







inline bool operator==(const ClsPointer& p1,const ClsPointer& p2){
	return p1.ID == p2.ID;}

inline bool operator!=(const ClsPointer& p1,const ClsPointer& p2){
	return p1.ID != p2.ID;}


/*void	StmpClsInfo::loadUpdate(const ClsPointer& p,const UserBit& ub){asrt(p.idx() < ldata.size());
	SharedClause* cls 	= ldata[p];
	uncount(cls);
	ldata[p]		= SharedClause::loadUpdate(cls,ub);
	count(ldata[p]);
	stamps[p]		= 0;
}*/

/*
void	ClsInfo::freeCls(const ClsPointer& p, const UserBit& ub,string why){
	asrt(cdata[p].cls != NULL);
	SharedClause* cls = cdata[p].cls;
	uint32_t sz = cls->countLits();
	freePtrOnly(p,sz);
	SharedClause::release(cls,ub,why);
}
*/


void	ClsInfo::freeCls(const ClsPointer& p, const UserBit& ub){
	asrt(cdata[p].cls != NULL);
	SharedClause* cls = cdata[p].cls;
	uint32_t sz = cls->countLits();
	freePtrOnly(p,sz);
	SharedClause::release(cls,ub);
}

void	ClsInfo::freePtrOnly(const ClsPointer& p,const uint32_t sz ){
	asrt(cdata[p].cls != NULL); asrt(sz != max32); asrt(cdata[p].cls->countLits() == sz);
	uncount(cdata[p].cls,sz);
	cdata[p].cls		= NULL;
	cdata[p].wtches		= free_data_head.idx();
	free_data_head 		= p;
	//used			.clear(p.idx());
}



ClsPointer	ClsInfo::nextValidBefore(uint32_t idx)const{
	asrt(idx <= cdata.size());
	while(idx--){
		if(cdata.byIndex(idx).cls !=NULL) return ClsPointer(idx);
	}
	return ClsPointer();
}

ClsPointer	ClsInfo::nextValidFrom(uint32_t idx)const{
	for(; idx < cdata.size();++idx){
		if(cdata.byIndex(idx).cls !=NULL) return ClsPointer(idx);
	}
	return ClsPointer();
}


ClsPointer ClsInfo::addCls(const Literal& w1, const Literal& w2, SharedClause* cls,const uint32_t sz,  const uint32_t& lbd){
	asrt(w1!=w2); asrt(cls != NULL); asrt(lbd <= SharedClause::LBD_MAX); asrt(cls->countLits() <= sz); asrt(cls->countLits() >= sz);
	ClsPointer ret = free_data_head;
	if(!ret.isNull()){		asrt(cdata.inRange(ret)) ;
		free_data_head 		= ClsPointer(cdata[ret].wtches);}
	else{
		ret = cdata			.grow(ClsWrp());}

	asrt(cdata.inRange(ret)) ;
	ClsWrp& cw 	= cdata[ret];
	cw.wtches 	= w1.idx() ^ w2.idx();
	cw.do_bup 	= 0;
	cw.lockd	= 0;
	cw.Lbd		= lbd * cls->isLearnt();
	cw.offset	= 0;
	cw.cls		= cls;


	count(cls,sz); asrt(cw.Lbd == 0|| cls->isLearnt());
	return ret;
}



uint32_t ClsInfo::getLBDWithRank(const uint32_t rank){

	std::vector<uint32_t>	lbd_cnt(SharedClause::LBD_MAX+2,0);

	for(uint32_t i = cdata.size();i--;){
		ClsPointer clp(i);
		ClsWrp& cw = cdata[clp];
		if(cw.lockd){	continue;}

		SharedClause* cls 	= cw.cls; 		
		if(cls == NULL || !cls->isLearnt() || cls->isWaste()){continue;}

		lbd_cnt[cw.getLBD()]++;
	}

	uint32_t take = 1;
	for(uint32_t tot = 0;tot <= rank && take < lbd_cnt.size() ;++take){
		tot += lbd_cnt[take];
	}

	LOG1("learntsGC max lbd take: " << take );
	return take;
}





/*
uint32_t ClsInfo::getLBDWithRank(const uint32_t rank){
	uint32_t buffer_sz 	= rank << 1;
	vector<uint32_t> 	lbds;
	lbds				.reserve(buffer_sz);
	uint32_t max_take 	= SharedClause::LBD_MAX; asrt(SharedClause::LBD_MAX>0);
	uint32_t max_reach	= 0;

	for(uint32_t i = cdata.size();i--;){
		ClsPointer clp(i);
		ClsWrp& cw = cdata[clp];
		if(cw.lockd){	continue;}

		SharedClause* cls 	= cw.cls;
		if(cls == NULL || !cls->isLearnt() || cls->isWaste()){continue;}

		uint32_t curr_lbd 	= cw.getLBD();
		if(curr_lbd>max_take) continue;

		lbds.push_back(curr_lbd); asrt(cls->isLearnt());
		if(curr_lbd > max_reach){max_reach = curr_lbd;}

		if(lbds.size() >= buffer_sz){
			LOG3("n_th element in between with rank "<<rank);
			std::nth_element(lbds.begin(), lbds.begin() + rank,lbds.end()); asrt(lbds[0]<=lbds[rank]);
			max_take	= lbds[rank];
			lbds.erase(lbds.begin() +rank, lbds.end());
		}
	}

	if(rank < lbds.size()){
		LOG3("n_th element at end with rank"<<rank);
		std::nth_element(lbds.begin(), lbds.begin() + rank,lbds.end()); asrt(lbds[0]<=lbds[rank]);
		max_take = lbds[rank];
	}
	LOG2("learntsGC max take: " << max_take << " max_reach: " << max_reach);
	return max_take < max_reach ? max_take : SharedClause::LBD_MAX;
}
*/











#endif /* CLAUSESTORAGE_HPP_ */
