/*
 * Logging.hpp
 *
 *  Created on: 13.08.2009
 *      Author: stephan
 */

#ifndef LOGGING_HPP_
#define LOGGING_HPP_

#include <stdio.h>
#include <stdarg.h>
#include <vector>
#include <string>
#include <omp.h>
#include <fstream>

using std::vector;
using std::string;



/* Choose one
 * Higher level implies lower levels
 * LOGGING_ENABLED will be defined if at least one logging type is defined
 */
//#define LOGGING_DETAILS
//#define LOGGING_MEDIUM
//#define LOGGING_BASIC

//for debugging: print currently important messages
//#define LOGGING_ACTUAL


#define SUPRESS_FIRST  0
/* LOGGING_DETAILS enables function LOG_DETAIL
 * LOGGING_MEDIUM enables function LOG_MEDIUM
 * LOGGING_BASIC enables function LOG_BASIC
 */


//from here it's automatic

//#define CONSOLE(args) std::cerr << "Thread " <<  omp_get_thread_num() << ": "<< args ;


inline double cpuSysTime(){
	struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_stime.tv_sec + (double)ru.ru_stime.tv_usec / 1000000.0;
};


inline double cpuUUUTime(){
	struct rusage ru;
    getrusage(RUSAGE_SELF, &ru);
    return (double)ru.ru_utime.tv_sec*1000000.0 + (double)ru.ru_utime.tv_usec ;
};

inline double cpuUsrTime(){return cpuUUUTime() / 1000000.0;};

struct Logging{
	static vector<std::ofstream*> 	outs;
	static vector<uint64_t> 		counts;


#if defined (LOGGING_DETAILS) or defined (LOGGING_MEDIUM) or defined (LOGGING_BASIC) or defined (LOGGING_ACTUAL)
	static void setNumberThreads(const uint32_t n,string path){
		assert(outs.empty());
		char buf[30];
		for(uint i = 0; i != n; ++i){
			sprintf(buf,"%slog_thread%02d.csv",path.c_str(),i);
			//outs.push_back(fopen (buf, "w+"));
			outs	.push_back(new std::ofstream(buf));
			counts	.push_back(0);
		}
		assert(outs.size()==n); assert(counts.size()==n);
	}

	static std::ofstream& getOut(){
		uint32_t thread = omp_get_thread_num();
		assert(thread < outs.size()); assert(outs[thread] != NULL);
		std::ofstream& 	out = *outs[thread]; assert(!out.bad());

		out << std::endl << cpuUUUTime() << " & " << thread << " & ";
		return out;
	}

	static uint64_t& getCount(){
		uint32_t thread = omp_get_thread_num();
		assert(thread < counts.size());
		return counts[thread];}

	static bool canLog(){
		return getCount()++ >= SUPRESS_FIRST;}


	static void closeThreads(){
		for(uint i = outs.size();i--;){	if(outs[i] != NULL) delete outs[i];}
		outs.clear();
	}


#ifndef LOGGING_ENABLED
#define LOGGING_ENABLED
#endif

#else
	static void setNumberThreads(const uint32_t n,string path){}
	//static void write(const char* format, ... ){}
	static void closeThreads(){}
#endif
};



#ifdef LOGGING_DETAILS

#define LOG3(args) if(Logging::canLog()){Logging::getOut() << args <<";" << std::flush;}

#ifndef LOGGING_MEDIUM
#define LOGGING_MEDIUM
#endif


#else

#define LOG3(args) ( assert(true))

#endif



#ifdef LOGGING_MEDIUM

#define LOG2(args) if(Logging::canLog()){Logging::getOut() << args <<";"<< std::flush ;}


#ifndef LOGGING_BASIC
#define LOGGING_BASIC
#endif

#else

#define LOG2(args) ( assert(true))

#endif



#ifdef LOGGING_BASIC

#define LOG1(args) if(Logging::canLog()){Logging::getOut() << args <<";"<< std::flush;}
#define LOG_COUNT 		Logging::getOut() << "Number of logs " << Logging::getCount() << std::flush;

#else

#define LOG1(args) ( assert(true))
#define LOG_COUNT 		( assert(true))

#endif

#ifdef LOGGING_ACTUAL
#define LOG0(args) {Logging::getOut() << args <<";"<< std::flush;} //cout << "c "<< args << endl;}
#endif

//#ifndef NDEBUG
#define ASRT_PRNT(args) std::cerr << "Thread " <<  omp_get_thread_num() << ": "<< args ; LOG1(args);
//#else
//#define ASRT_PRNT(args) std::cerr << endl << "c DONT USE THIS PRINT FUNCTION OUTSIDE ASSERTIONS" << endl;
//#endif

#define logging_hpp_version "$Revision: 1.7 $"


#endif /* LOGGING_HPP_ */
